#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BSD 3-Clause License
#
# Copyright 2022, by the California Institute of Technology.
# ALL RIGHTS RESERVED. United States Government Sponsorship acknowledged.
# Any commercial use must be negotiated with the Office of Technology Transfer
# at the California Institute of Technology.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Random number generators.

Authors:
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
    Natalia Korsakova <natalia.korsakova@obspm.fr>
"""

import numpy as np


def poisson_interarrival(lam=1.0, size=None):
    """Draws arrival time differences for events following a Poisson distribution.

    Args:
        lam (float): expected number of events occurring per second [/s]
        size (int): output shape, None for a single scalar value
    """
    return np.random.exponential(1 / lam, size)


def poisson_arrival(lam=1.0, t0=0.0, size=None):
    """Draws arrival times for events following a Poisson distribution.

    Args:
        lam (float): expected number of events occurring per second [/s]
        t0 (float): random process initial time [s]
        size (int): output shape, None for a single scalar value
    """
    cumsum = np.cumsum(poisson_interarrival(lam, size))
    if size is None:
        cumsum = cumsum.item()
    return t0 + cumsum



"""
Code borrowed from https://github.com/bayesiains/nflows
"""

import torch
from torch import nn
from torch.nn import functional as F
from torch.nn import init


class ResidualBlock(nn.Module):
    """A general-purpose residual block. Works only with 1-dim inputs."""

    def __init__(
        self,
        features,
        context_features,
        activation=F.relu,
        dropout_probability=0.0,
        use_batch_norm=False,
        zero_initialization=True,
    ):
        super().__init__()
        self.activation = activation

        self.use_batch_norm = use_batch_norm
        if use_batch_norm:
            self.batch_norm_layers = nn.ModuleList(
                [nn.BatchNorm1d(features, eps=1e-3) for _ in range(2)]
            )
        if context_features is not None:
            self.context_layer = nn.Linear(context_features, features)
        self.linear_layers = nn.ModuleList(
            [nn.Linear(features, features) for _ in range(2)]
        )
        self.dropout = nn.Dropout(p=dropout_probability)
        if zero_initialization:
            init.uniform_(self.linear_layers[-1].weight, -1e-3, 1e-3)
            init.uniform_(self.linear_layers[-1].bias, -1e-3, 1e-3)

    def forward(self, inputs, context=None):
        """
        Forward pass of ResidualBlock.
        """
        temps = inputs
        if self.use_batch_norm:
            temps = self.batch_norm_layers[0](temps)
        temps = self.activation(temps)
        temps = self.linear_layers[0](temps)
        if self.use_batch_norm:
            temps = self.batch_norm_layers[1](temps)
        temps = self.activation(temps)
        temps = self.dropout(temps)
        temps = self.linear_layers[1](temps)
        if context is not None:
            temps = F.glu(torch.cat((temps, self.context_layer(context)), dim=1), dim=1)
        return inputs + temps


class ResidualNet(nn.Module):
    """A general-purpose residual network. Works only with 1-dim inputs."""

    def __init__(
        self,
        in_features,
        out_features,
        hidden_features,
        context_features=None,
        num_blocks=15,
        activation=F.elu,
        dropout_probability=0.0,
        use_batch_norm=False,
    ):
        super().__init__()
        self.hidden_features = hidden_features
        self.context_features = context_features
        if context_features is not None:
            self.initial_layer = nn.Linear(
                in_features + context_features, hidden_features
            )
        else:
            self.initial_layer = nn.Linear(in_features, hidden_features)
        self.blocks = nn.ModuleList(
            [
                ResidualBlock(
                    features=hidden_features,
                    context_features=context_features,
                    activation=activation,
                    dropout_probability=dropout_probability,
                    use_batch_norm=use_batch_norm,
                )
                for _ in range(num_blocks)
            ]
        )
        self.final_layer = nn.Linear(hidden_features, out_features)

    def forward(self, inputs, context=None):
        """
        Forward pass.
        """
        if context is None:
            temps = self.initial_layer(inputs)
        else:
            temps = self.initial_layer(torch.cat((inputs, context), dim=1))
        for block in self.blocks:
            temps = block(temps, context=context)
        outputs = self.final_layer(temps)
        return outputs


class ConvResidualBlock(nn.Module):
    """
    Convolutional Residual Block.
    """

    def __init__(
        self,
        channels,
        context_channels=None,
        activation=F.relu,
        dropout_probability=0.0,
        use_batch_norm=False,
        zero_initialization=True,
    ):
        super().__init__()
        self.activation = activation

        if context_channels is not None:
            self.context_layer = nn.Conv2d(
                in_channels=context_channels,
                out_channels=channels,
                kernel_size=1,
                padding=0,
            )
        self.use_batch_norm = use_batch_norm
        if use_batch_norm:
            self.batch_norm_layers = nn.ModuleList(
                [nn.BatchNorm2d(channels, eps=1e-3) for _ in range(2)]
            )
        self.conv_layers = nn.ModuleList(
            [nn.Conv2d(channels, channels, kernel_size=3, padding=1) for _ in range(2)]
        )
        self.dropout = nn.Dropout(p=dropout_probability)
        if zero_initialization:
            init.uniform_(self.conv_layers[-1].weight, -1e-3, 1e-3)
            init.uniform_(self.conv_layers[-1].bias, -1e-3, 1e-3)

    def forward(self, inputs, context=None):
        """
        Forward pass of Convolutional Residual Block.
        """
        temps = inputs
        if self.use_batch_norm:
            temps = self.batch_norm_layers[0](temps)
        temps = self.activation(temps)
        temps = self.conv_layers[0](temps)
        if self.use_batch_norm:
            temps = self.batch_norm_layers[1](temps)
        temps = self.activation(temps)
        temps = self.dropout(temps)
        temps = self.conv_layers[1](temps)
        if context is not None:
            temps = F.glu(torch.cat((temps, self.context_layer(context)), dim=1), dim=1)
        return inputs + temps


class ConvResidualNet(nn.Module):
    """
    Network constructed of convolutional residual blocks.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        hidden_channels,
        context_channels=None,
        num_blocks=2,
        activation=F.relu,
        dropout_probability=0.0,
        use_batch_norm=False,
    ):
        super().__init__()
        self.context_channels = context_channels
        self.hidden_channels = hidden_channels
        if context_channels is not None:
            self.initial_layer = nn.Conv2d(
                in_channels=in_channels + context_channels,
                out_channels=hidden_channels,
                kernel_size=1,
                padding=0,
            )
        else:
            self.initial_layer = nn.Conv2d(
                in_channels=in_channels,
                out_channels=hidden_channels,
                kernel_size=1,
                padding=0,
            )
        self.blocks = nn.ModuleList(
            [
                ConvResidualBlock(
                    channels=hidden_channels,
                    context_channels=context_channels,
                    activation=activation,
                    dropout_probability=dropout_probability,
                    use_batch_norm=use_batch_norm,
                )
                for _ in range(num_blocks)
            ]
        )
        self.final_layer = nn.Conv2d(
            hidden_channels, out_channels, kernel_size=1, padding=0
        )

    def forward(self, inputs, context=None):
        """
        Forward pass of residual network.
        """
        if context is None:
            temps = self.initial_layer(inputs)
        else:
            temps = self.initial_layer(torch.cat((inputs, context), dim=1))
        for block in self.blocks:
            temps = block(temps, context)
        outputs = self.final_layer(temps)
        return outputs



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BSD 3-Clause License
#
# Copyright 2022, by the California Institute of Technology.
# ALL RIGHTS RESERVED. United States Government Sponsorship acknowledged.
# Any commercial use must be negotiated with the Office of Technology Transfer
# at the California Institute of Technology.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Gap mask generation module for LISA mission data processing.

This module provides the GapMaskGenerator class for creating realistic data gap masks
that simulate various types of mission interruptions including planned maintenance,
communication blackouts, and unplanned outages.

Authors:
    Ollie Burke <ollie.burke@glasgow.ac.uk>
    Eleonora Castelli <eleonora.castelli@unitn.it>
"""

import json
from pathlib import Path
from typing import Any, Union

import h5py
import numpy as np
from lisaconstants import TROPICALYEAR_J2000DAY
from numpy.typing import NDArray
from scipy.stats import expon

SECONDS_PER_YEAR = TROPICALYEAR_J2000DAY * 86400
AN_HOUR = 60 * 60


class GapMaskGenerator:
    """
    A class to generate and manage gap masks for time series data. Original code developed by
    Eleonora Castelli (NASA Goddard) and adapted by Ollie Burke (Glasgow)
    """

    def __init__(
        self,
        sim_t: NDArray[np.floating[Any]],
        gap_definitions: dict[str, dict[str, dict[str, Any]]],
        treat_as_nan: bool = True,
        planseed: int = 11071993,
        unplanseed: int = 16121997,
    ):
        """
        Parameters
        ----------
        sim_t : np.ndarray
            Array of simulation time values.
        gap_definitions : dict
            Dictionary defining planned and unplanned gaps. Each gap entry must include:
            - 'rate_per_year' (float)
            - 'duration_hr' (float)
        treat_as_nan : bool, optional
            If True, gaps are inserted as NaNs. If False, they are inserted as zeros.
        planseed : int
            Seed for planned gap randomization.
        unplanseed : int
            Seed for unplanned gap randomization.
        """

        # Initialise sampling properties
        self.sim_t = sim_t
        dt = sim_t[1] - sim_t[0]
        self.dt = dt
        self.n_data = len(sim_t)

        # Determine whether or not mask uses nans or zeros
        self.treat_as_nan = treat_as_nan

        # Set seeds
        self.planseed = planseed
        self.unplanseed = unplanseed

        # Read in gap definitions, rates, durations, types.
        self.gap_definitions: dict[str, dict[str, dict[str, float]]] = {}

        # Quick and simple error checking
        for kind in ("planned", "unplanned"):
            if kind not in gap_definitions:
                raise ValueError(f"Missing '{kind}' section in gap_definitions.")

            self.gap_definitions[kind] = {}

            for name, entry in gap_definitions[kind].items():
                if "rate_per_year" not in entry or "duration_hr" not in entry:
                    raise ValueError(
                        f"Gap '{name}' in '{kind}' must have 'rate_per_year' and 'duration_hr'."
                    )
                if entry["rate_per_year"] < 0 or entry["duration_hr"] < 0:
                    raise ValueError(
                        f"Gap '{name}' in '{kind}' has negative rate or duration."
                    )

                self.gap_definitions[kind][name] = {
                    "rate_per_year": entry["rate_per_year"],
                    "duration_hr": entry["duration_hr"],
                }
        # If the gap definitions pass (not empty, negative etc.)
        # then we can set the gap definitions
        self._update_gap_arrays()

    def _update_gap_arrays(self) -> None:
        """
        Helper function to update the gap arrays based on the gap definitions.
        This function is called during initialization and when gap definitions are updated.

        We convert the rate_per_year into rate_per_second and duration_hr into samples
        """

        # Extract gap types for planned/unplanned gaps
        self.planned_labels = list(self.gap_definitions["planned"].keys())
        self.unplanned_labels = list(self.gap_definitions["unplanned"].keys())

        self.planned_rates = (
            np.array(
                [
                    self.gap_definitions["planned"][k]["rate_per_year"]
                    / SECONDS_PER_YEAR
                    for k in self.planned_labels
                ]
            )
            * self.dt
        )
        self.planned_durations = (
            np.array(
                [
                    self.gap_definitions["planned"][k]["duration_hr"] * 3600
                    for k in self.planned_labels
                ]
            )
            / self.dt
        )

        self.unplanned_rates = (
            np.array(
                [
                    self.gap_definitions["unplanned"][k]["rate_per_year"]
                    / SECONDS_PER_YEAR
                    for k in self.unplanned_labels
                ]
            )
            * self.dt
        )
        self.unplanned_durations = (
            np.array(
                [
                    self.gap_definitions["unplanned"][k]["duration_hr"] * 3600
                    for k in self.unplanned_labels
                ]
            )
            / self.dt
        )

    def construct_planned_gap_mask(
        self,
        rate: float,
        gap_duration: float,
        seed: Union[int, None] = None,
    ) -> NDArray[Union[np.float64, np.int_]]:
        """
        Construct a planned gap mask with regular spacing and jitter.

        Parameters
        ----------
        rate : float
            Gap rate (in events/s).
        gap_duration : float
            Gap duration (in samples).
        seed : int or None
            Random seed.

        Returns
        -------
        np.ndarray
            Array with gaps represented as NaNs or zeros, but valid data always as integer 1.
        """
        # Set specific seed
        np.random.seed(seed)
        if self.treat_as_nan:
            mask = np.ones(self.n_data, dtype=np.float64)
        else:
            mask = np.ones(self.n_data, dtype=int)
        est_num_gaps = int(self.n_data * rate)

        jitter = 0.2 * gap_duration * (np.random.rand(est_num_gaps) - 0.5)
        gap_starts = (jitter + (1 / rate) * np.arange(1, est_num_gaps + 1)).astype(int)
        gap_ends = (gap_starts + gap_duration).astype(int)

        # Decide whether or not to treat gaps as NaNs or 0s.
        gap_val = self._gap_value()

        for start, end in zip(gap_starts, gap_ends):
            if end < self.n_data:
                mask[start:end] = gap_val

        return mask

    def construct_unplanned_gap_mask(
        self,
        rate: float,
        gap_duration: float,
        seed: Union[int, None] = None,
    ) -> NDArray[Union[np.float64, np.int_]]:
        """
        Construct an unplanned gap mask using an exponential distribution.

        Parameters
        ----------
        rate : float
            Gap rate (in events/s).
        gap_duration : float
            Gap duration (in seconds).
        seed : int or None
            Random seed.

        Returns
        -------
        np.ndarray
            Array with gaps represented as NaNs or zeros, but valid data always as integer 1.
        """
        np.random.seed(seed)
        if self.treat_as_nan:
            mask = np.ones(self.n_data, dtype=np.float64)
        else:
            mask = np.ones(self.n_data, dtype=int)

        est_num_gaps = int(self.n_data * rate)
        if est_num_gaps == 0 and np.random.rand() < rate * self.n_data:
            est_num_gaps = 1

        start_offsets = expon.rvs(scale=1 / rate, size=est_num_gaps).astype(int)
        gap_starts = np.cumsum(start_offsets)
        gap_starts = gap_starts[gap_starts + gap_duration < self.n_data]
        gap_ends = (gap_starts + gap_duration).astype(int)

        # Decide whether or not to treat gaps as NaNs or 0s.
        gap_val = self._gap_value()

        for start, end in zip(gap_starts, gap_ends):
            mask[start:end] = gap_val

        return mask

    def _gap_value(self) -> Union[float, int]:
        """
        Helper function to determine the value used for gaps in the mask.
        Returns NaN if treat_as_nan is True, otherwise returns 0.
        """
        return np.nan if self.treat_as_nan else 0

    def generate_mask(
        self,
        include_planned: bool = True,
        include_unplanned: bool = True,
    ) -> NDArray[Union[np.float64, np.int_]]:
        """
        Combine planned and unplanned masks into a final mask.

        Parameters
        ----------
        include_planned : bool
            Include planned gaps.
        include_unplanned : bool
            Include unplanned gaps.

        Returns
        -------
        np.ndarray
            Final gap mask with valid data as integer 1, gaps as 0 or NaN.
        """
        if self.treat_as_nan:
            mask = np.ones(self.n_data, dtype=np.float64)
        else:
            mask = np.ones(self.n_data, dtype=int)

        # Construct planned gap mask
        if include_planned:
            for rate, duration in zip(self.planned_rates, self.planned_durations):
                planned_mask = self.construct_planned_gap_mask(
                    rate, duration, seed=self.planseed
                )
                mask = mask * planned_mask.astype(mask.dtype)

        # Construct unplanned gap mask
        if include_unplanned:
            for rate, duration in zip(self.unplanned_rates, self.unplanned_durations):
                unplanned_mask = self.construct_unplanned_gap_mask(
                    rate, duration, seed=self.unplanseed
                )
                mask = mask * unplanned_mask.astype(mask.dtype)

        return mask

    def save_to_hdf5(
        self,
        mask: np.ndarray,
        filename: str = "gap_mask_data.h5",
        save_as_quality_flags: bool = False,
    ) -> None:
        """
        Save the gap mask and associated simulation metadata to an HDF5 file.

        Parameters
        ----------
        mask : np.ndarray
            The gap mask array, typically generated using `generate_mask()`.
            Should be of the same length as `sim_t`, and contain either 1s and 0s,
            or 1s and NaNs depending on the `treat_as_nan` setting.

        filename : str, optional
            Path to the HDF5 file to create. Defaults to "gap_mask_data.h5".

        save_as_quality_flags : bool, optional
            If True, converts the mask to quality data flags before saving:
            - 1 = gap/corrupt data (where original mask has NaN or 0)
            - 0 = valid data (where original mask has 1.0 or 1)
            If False (default), saves the mask in its original format.

        Notes
        -----
        This function stores:

        - The mask data under `"gap_mask"` (original format) or `"quality_flags"` (if save_as_quality_flags=True)
        - Metadata attributes:

          - `"dt"` (time step)
          - `"treat_as_nan"` (boolean mask type flag)
          - `"saved_as_quality_flags"` (boolean indicating storage format)

        - Gap configuration details in two groups:

          - `"planned_gaps"`: each with `rate_events_per_year` and `duration_hours`
          - `"unplanned_gaps"`: same structure as planned gaps

        The resulting file can be reloaded using the `from_hdf5()` class method given below.

        """
        with h5py.File(filename, "w") as f:
            # Convert mask to quality flags if requested
            if save_as_quality_flags:
                # Convert to quality flags: 1 = gap/corrupt, 0 = valid
                if self.treat_as_nan:
                    # Input mask: 1.0 = valid, NaN = gap
                    # Output flags: 0 = valid, 1 = gap
                    quality_flags = np.where(np.isnan(mask), 1, 0).astype(np.int32)
                else:
                    # Input mask: 1 = valid, 0 = gap
                    # Output flags: 0 = valid, 1 = gap
                    quality_flags = np.where(mask == 0, 1, 0).astype(np.int32)

                f.create_dataset("quality_flags", data=quality_flags, compression="lzf")
            else:
                # Save original mask format
                f.create_dataset("gap_mask", data=mask, compression="lzf")

            # Save simulation time metadata as (t_0, t_end, dt)
            f.attrs["sim_time_info -- [t0, t_end, dt]"] = (
                float(self.sim_t[0]),
                float(self.sim_t[-1]),
                float(self.dt),
            )
            f.attrs["treat_as_nan"] = self.treat_as_nan
            f.attrs["saved_as_quality_flags"] = save_as_quality_flags

            # Save planned gap info
            planned_grp = f.create_group("planned_gaps")
            for label in self.planned_labels:
                grp = planned_grp.create_group(label)
                grp.attrs["rate_events_per_year"] = self.gap_definitions["planned"][
                    label
                ]["rate_per_year"]
                grp.attrs["duration_hours"] = self.gap_definitions["planned"][label][
                    "duration_hr"
                ]

            # Save unplanned gap info
            unplanned_grp = f.create_group("unplanned_gaps")
            for label in self.unplanned_labels:
                grp = unplanned_grp.create_group(label)
                grp.attrs["rate_events_per_year"] = self.gap_definitions["unplanned"][
                    label
                ]["rate_per_year"]
                grp.attrs["duration_hours"] = self.gap_definitions["unplanned"][label][
                    "duration_hr"
                ]

    @classmethod
    def from_hdf5(cls, filename: str) -> "GapMaskGenerator":
        """
        Reconstruct a GapMaskGenerator object from an HDF5 file.
        classmethod, so no need to instantiate the class first.
        This method reads the gap mask, simulation time, and metadata from the file,
        and returns a new instance of GapMaskGenerator.

        Parameters
        ----------
        filename : str
            Path to the HDF5 file.

        Returns
        -------
        GapMaskGenerator
            A new instance reconstructed from the file.
        """

        with h5py.File(filename, "r") as f:
            # Store simulation time, sampling interval, and mask type.
            sim_time_info = f.attrs["sim_time_info -- [t0, t_end, dt]"]
            sim_time_array = np.array(sim_time_info)  # Ensure it's a numpy array
            t0, t_end, dt = (
                float(sim_time_array[0]),
                float(sim_time_array[1]),
                float(sim_time_array[2]),
            )
            sim_t = np.arange(t0, t_end + dt, dt)
            treat_as_nan = bool(
                f.attrs.get("treat_as_nan", True)
            )  # fallback True if not saved

            # Store information about the planned gaps
            planned_gaps = {}
            for key in f["planned_gaps"]:
                grp = f["planned_gaps"][key]
                planned_gaps[key] = {
                    "rate_per_year": float(grp.attrs["rate_events_per_year"]),
                    "duration_hr": float(grp.attrs["duration_hours"]),
                }

            # Store information about the unplanned gaps
            unplanned_gaps = {}
            for key in f["unplanned_gaps"]:
                grp = f["unplanned_gaps"][key]
                unplanned_gaps[key] = {
                    "rate_per_year": float(grp.attrs["rate_events_per_year"]),
                    "duration_hr": float(grp.attrs["duration_hours"]),
                }

            gap_definitions = {
                "planned": planned_gaps,
                "unplanned": unplanned_gaps,
            }
            # Return as class object
        return cls(
            sim_t=sim_t,
            gap_definitions=gap_definitions,
            treat_as_nan=treat_as_nan,
        )

    @staticmethod
    def load_mask_from_hdf5(
        filename: str, convert_to_mask: bool = True
    ) -> tuple[np.ndarray, dict[str, Any]]:
        """
        Load just the mask data from an HDF5 file, with optional conversion.

        Parameters
        ----------
        filename : str
            Path to the HDF5 file.
        convert_to_mask : bool, optional
            If True and the file contains quality flags, convert them back to a mask format.
            If False, return the data in its stored format.

        Returns
        -------
        tuple[np.ndarray, dict]
            The mask/quality flag data and metadata dictionary containing:
            - 'treat_as_nan': boolean indicating original mask type
            - 'saved_as_quality_flags': boolean indicating storage format
            - 'data_type': string describing what was returned

        Examples
        --------
        >>> # Load quality flags as-is
        >>> flags, meta = GapMaskGenerator.load_mask_from_hdf5("data.h5", convert_to_mask=False)
        >>> # Load and convert quality flags to mask format
        >>> mask, meta = GapMaskGenerator.load_mask_from_hdf5("data.h5", convert_to_mask=True)
        """
        with h5py.File(filename, "r") as f:
            treat_as_nan = bool(f.attrs.get("treat_as_nan", True))
            saved_as_quality_flags = bool(f.attrs.get("saved_as_quality_flags", False))

            metadata = {
                "treat_as_nan": treat_as_nan,
                "saved_as_quality_flags": saved_as_quality_flags,
                "data_type": "",  # Will be set below
            }

            if saved_as_quality_flags:
                # File contains quality flags (1=gap, 0=valid)
                quality_flags = f["quality_flags"][:]

                if convert_to_mask:
                    # Convert quality flags back to mask format
                    if treat_as_nan:
                        # Convert to NaN mask: 0->1.0 (valid), 1->NaN (gap)
                        mask_data = np.where(quality_flags == 0, 1.0, np.nan)
                        metadata["data_type"] = "converted_to_nan_mask"
                    else:
                        # Convert to binary mask: 0->1 (valid), 1->0 (gap)
                        mask_data = np.where(quality_flags == 0, 1, 0)
                        metadata["data_type"] = "converted_to_binary_mask"
                    return mask_data, metadata

                metadata["data_type"] = "quality_flags"
                return quality_flags, metadata

            # File contains original mask format
            mask_data = f["gap_mask"][:]
            metadata["data_type"] = "original_mask"
            return mask_data, metadata

    def summary(
        self,
        mask: Union[NDArray[np.float64], NDArray[np.int_], None] = None,
        export_json_path: Union[str, Path, None] = None,
    ) -> dict[str, Any]:
        """
        Return a structured dictionary summarising the gap configuration
        and optionally the content of a specific mask.

        Parameters
        ----------
        mask : np.ndarray, optional
            If provided, calculates duty cycle and number of gaps based on this mask.

        export_json_path : str or Path, optional
            If provided, writes the summary dictionary to a JSON file at the given path.

        Returns
        -------
        dict
            Summary of configuration and optionally mask content.
        """
        # Extract as much information as possible from the gap definitions
        # and the mask if provided.
        summary_dict: dict[str, Any] = {
            "simulation": {
                "dt": self.dt,
                "n_data": self.n_data,
                "duration_sec": self.n_data * self.dt,
                "duration_days": self.n_data * self.dt / 86400,
            },
            "seeds": {
                "planned": self.planseed,
                "unplanned": self.unplanseed,
            },
            "planned_gaps": {},
            "unplanned_gaps": {},
        }

        for kind in ("planned", "unplanned"):
            for name, info in self.gap_definitions[kind].items():
                duration_sec = info["duration_hr"] * 3600
                duration_samples = int(duration_sec / self.dt)
                rate_per_sec = info["rate_per_year"] / SECONDS_PER_YEAR

                summary_dict[f"{kind}_gaps"][name] = {
                    "rate_events_per_year": info["rate_per_year"],
                    "duration_hr": info["duration_hr"],
                    "rate_events_per_sec": rate_per_sec,
                    "duration_sec": duration_sec,
                    "duration_samples": duration_samples,
                }

        # Add dynamic information from mask if provided
        if mask is not None:
            is_gap = np.isnan(mask) if self.treat_as_nan else (mask == 0)
            duty_cycle = 1.0 - np.sum(is_gap) / len(mask)

            summary_dict["mask_analysis"] = {
                "duty_cycle_percent": round(100 * duty_cycle, 4),
                "total_gap_samples": int(np.sum(is_gap)),
                "total_gap_hours": round(np.sum(is_gap) * self.dt / 3600, 2),
            }

            # Optional: Count number of contiguous gaps
            gap_count = 0
            in_gap = False
            for val in is_gap:
                if val and not in_gap:
                    gap_count += 1
                    in_gap = True
                elif not val:
                    in_gap = False
            summary_dict["mask_analysis"]["number_of_gap_segments"] = gap_count

        if export_json_path is not None:
            with open(export_json_path, "w", encoding="utf-8") as f:
                json.dump(summary_dict, f, indent=4)
        return summary_dict

    def build_quality_flags(
        self, data_array: NDArray[np.float64], save_to_file: Union[str, None] = None
    ) -> NDArray[np.int_]:
        """
        Build a masking function based on the gap definitions and the provided data array.

        Parameters
        ----------
        data_array : np.ndarray
            The data array to be masked.
        save_to_file : str, optional
            If provided, saves the quality flags to an HDF5 file at the specified path.

        Returns
        -------
        np.ndarray
            A masking function that can be applied to the data array.
        """
        # Wherever nans appear, replace this with integer 1.
        # Else 0 for valid data product.
        mask = np.where(np.isnan(data_array), 1, 0).astype(np.int32)

        if save_to_file is not None:
            # Convert quality flags back to mask format for saving
            float_mask = self.quality_flags_to_mask(mask)
            self.save_to_hdf5(
                float_mask, filename=save_to_file, save_as_quality_flags=True
            )

        return mask

    @staticmethod
    def quality_flags_to_mask(quality_flags: NDArray[np.int_]) -> NDArray[np.float64]:
        """
        Convert integer quality flags to a float mask suitable for data multiplication.

        This utility function converts from the compact storage format (integers)
        to the working format (floats with NaNs) used for masking data arrays.

        Parameters
        ----------
        quality_flags : NDArray[np.int_]
            Integer array where:
            - 0 = valid data
            - 1 = corrupt/gap data

        Returns
        -------
        NDArray[np.float64]
            Float mask array where:
            - 1.0 = valid data (multiply by 1.0 = unchanged)
            - NaN = gap data (multiply by NaN = NaN)

        Examples
        --------
        >>> flags = np.array([0, 1, 0, 1, 0], dtype=int)
        >>> mask = GapMaskGenerator.quality_flags_to_mask(flags)
        >>> print(mask)
        [1.0, nan, 1.0, nan, 1.0]
        >>> data = np.array([10.0, 20.0, 30.0, 40.0, 50.0])
        >>> masked_data = data * mask
        >>> print(masked_data)
        [10.0, nan, 30.0, nan, 50.0]
        """
        # Convert 0->1.0 (valid data) and 1->NaN (corrupt data)
        return np.where(quality_flags == 0, 1.0, np.nan)

    def _match_gap_length_to_label(
        self, gap_length_samples: int
    ) -> Union[tuple[str, str], None]:
        """
        Match a gap length to one of the known planned/unplanned definitions.

        Parameters
        ----------
        gap_length_samples : int
            Length of the gap in samples.

        Returns
        -------
        tuple of (kind, label) or None
        """
        for kind in ("planned", "unplanned"):
            for label in self.gap_definitions[kind]:
                expected = int(
                    self.gap_definitions[kind][label]["duration_hr"] * 3600 / self.dt
                )
                if (
                    abs(gap_length_samples - expected) < 0.1 * expected
                ):  # allow 10% fuzz
                    return (kind, label)
        return None



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BSD 3-Clause License
#
# Copyright 2022, by the California Institute of Technology.
# ALL RIGHTS RESERVED. United States Government Sponsorship acknowledged.
# Any commercial use must be negotiated with the Office of Technology Transfer
# at the California Institute of Technology.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Glitches using data generated externally.

Authors:
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
    Eleonora Castelli <eleonora.castelli@unitn.it>
"""

import logging
import random
from typing import Any, List

import h5py
import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline

from .base import Glitch

logger = logging.getLogger(__name__)


class TimeSeriesGlitch(Glitch):
    """Glitch with signal from a Numpy array.

    To honor the glitch's sampling parameters, the input data may be resampled
    using spline interpolation. If you do not wish to interpolate, make sure to
    instantiate the glitch with sampling parameters matching your data.

    Args:
        t ((N,) array-like): times associated with ``tseries`` [s]
        tseries ((N,) array-like): glitch signal [injection point unit]
        interp_order (int): spline-interpolation order [one of 1, 2, 3, 4, 5]
        ext (str):
            extrapolation mode for elements out of range, see
            `scipy.interpolate.InterpolatedUnivariateSpline
            <https://docs.scipy.org/doc/scipy/reference/reference/generated/
            scipy.interpolate.InterpolatedUnivariateSpline.html>`_
        **kwargs: all other args from :class:`lisaglitch.Glitch`
    """

    def __init__(self, t, tseries, interp_order=1, ext="const", **kwargs) -> None:
        super().__init__(**kwargs)
        self.interp_order = int(interp_order)
        self.ext = str(ext)  #: Extrapolation mode for elements out of range.

        # Compute spline interpolation
        logger.info("Computing spline interpolation from time series")
        self.interpolant = InterpolatedUnivariateSpline(
            t + self.t_inj,
            tseries,
            k=self.interp_order,
            ext=self.ext,
            check_finite=True,
        )  #: Glitch signal interpolating function.

    def compute_signal(self, t) -> np.ndarray:
        return self.interpolant(t)


class HDF5Glitch(TimeSeriesGlitch):
    """Glitch with signal read from a HDF5 file.

    A glitch time-series is extracted from a dataset and injected at time
    ``t_inj``.

    Give the full path to a given dataset as ``node`` argument, or use the path
    to a group to pick a random dataset inside this group. Use the root ``/`` to
    pick a random dataset in the entire HDF5 file.

    To honor the glitch's sampling parameters, the input data may be resampled
    using spline interpolation. If you do not wish to interpolate, make sure to
    instantiate the glitch with sampling parameters matching your data.

    Args:
        path (str): path to HDF5 file
        node (str): path to dataset, or group to pick a random dataset
        exclude (list of str): datasets to exclude when picking a random dataset
    """

    def __init__(self, path, node="/", exclude=None, **kwargs) -> None:
        self.path = str(path)  #: Path to HDF5 file.
        logger.info("Opening HDF5 file '%s'", self.path)
        hdf5 = h5py.File(self.path, "r")

        # Exclude time dataset by default
        if exclude is None:
            exclude = ["t"]
        self.exclude = exclude

        # Pick dataset
        self.node = str(node)
        group = hdf5[self.node]
        if isinstance(group, h5py.Group):
            logger.debug("Picking a random dataset in group '%s'", self.node)
            dataset_names: List[str] = []
            group.visititems(
                lambda _, node: self._append_dataset_name(node, dataset_names)
            )
            self.dataset = random.choice(dataset_names)
            while self.dataset in self.exclude:
                self.dataset = random.choice(dataset_names)
        elif isinstance(group, h5py.Dataset):
            logger.debug("Using user-provided dataset '%s'", self.node)
            self.dataset = self.node  #: Glitch dataset.

        # Read dataset and time vector
        logger.info("Reading dataset '%s'", self.dataset)
        dataset = hdf5[self.dataset]

        hdf5_dt = float(self._get_closest_attr("dt", dataset))
        try:
            hdf5_t0 = float(self._get_closest_attr("t0", dataset))
        except KeyError:
            hdf5_t0 = 0
        try:
            hdf5_size = int(self._get_closest_attr("size", dataset))
        except KeyError:
            hdf5_size = len(dataset)

        hdf5_t = hdf5_t0 + np.arange(hdf5_size, dtype=np.float64) * hdf5_dt
        super().__init__(hdf5_t, dataset[:], **kwargs)

        logger.info("Closing HDF5 file '%s'", self.path)
        hdf5.close()

    @staticmethod
    def _append_dataset_name(node, datasets) -> None:
        """Append node's name to ``datasets`` if node is a dataset.

        Use this method with :meth:`h5py.Group.visititems` to list datasets in a
        group.

        Args:
            node (str): HDF5 node, either a dataset or a group
            datasets (list of str): list on which to append the dataset's name
        """
        if isinstance(node, h5py.Dataset):
            datasets.append(node.name)

    @staticmethod
    def _get_closest_attr(attr, node) -> Any:
        """Get value of attribute ``attr`` of the nearest ancestor of ``node``.

        We try to return the value of attribute ``attr`` on the node; if it does
        not exist, we go up the file hierarchy and try to read and return the
        attribute on the nearest ancestor.

        Args:
            node (str): a group or a dataset in an HDF5 file

        Raises:
            ``KeyError`` if no ancestor defines the attribute ``attr``.
        """
        if attr in node.attrs:
            return node.attrs[attr]
        if node.parent == node:
            raise KeyError(
                f"the attribute '{attr}' is not defined on '{node.file.filename}'"
            )
        return HDF5Glitch._get_closest_attr(attr, node.parent)

    def plot(self, *args, **kwargs) -> None:
        """Plot glitch signal.

        See `Glitch.plot()`.
        """
        if "title" not in kwargs:
            kwargs["title"] = f"Glitch from dataset '{self.dataset}' in '{self.path}'"
        super().plot(*args, **kwargs)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BSD 3-Clause License
#
# Copyright 2022, by the California Institute of Technology.
# ALL RIGHTS RESERVED. United States Government Sponsorship acknowledged.
# Any commercial use must be negotiated with the Office of Technology Transfer
# at the California Institute of Technology.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Mathematical glitches.

Defines glitches based on simple mathematical functions.

Authors:
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
    Eleonora Castelli <eleonora.castelli@unitn.it>
"""

import logging

import numpy as np

from .base import Glitch

logger = logging.getLogger(__name__)


class FunctionGlitch(Glitch):
    """Glitch with signal from a Python function.

    The function will be called with the time array as argument.

    Args:
        func (callable): function to compute the glitch signal, taking a time
            array as argument and returning an array of the same length
            [injection point unit]
        **kwargs: all other args from :class:`lisaglitch.Glitch`
    """

    def __init__(self, func, **kwargs) -> None:
        super().__init__(**kwargs)
        self.func = func

    def compute_signal(self, t) -> np.ndarray:
        return self.func(t)


class StepGlitch(Glitch):
    """Step glitch.

    A step glitch is vanishing before ``t_inj``, and is ``level`` after.

    Args:
        level (float): amplitude [injection point unit]
        **kwargs: all other args from :class:`lisaglitch.Glitch`
    """

    def __init__(self, level=1, **kwargs) -> None:
        super().__init__(**kwargs)
        self.level = float(level)  #: Amplitude [injection point unit].

    def compute_signal(self, t) -> np.ndarray:
        return np.where(t >= self.t_inj, self.level, 0)


class RectangleGlitch(Glitch):
    """Rectangular glitch.

    A rectangular glitch is vanishing except between ``t_inj`` and ``t_inj +
    width``, where it is ``level``.

    Args:
        width (float): width of the rectangle (glitch duration) [s]
        level (float): amplitude [injection point unit]
        **kwargs: all other args from :class:`lisaglitch.Glitch`
    """

    def __init__(self, width, level=1, **kwargs) -> None:
        super().__init__(**kwargs)
        self.width = float(width)  #: Glitch duration) [s].
        self.level = float(level)  #: Amplitude [injection point unit].

    def compute_signal(self, t) -> np.ndarray:
        inside = np.logical_and(t >= self.t_inj, t < self.t_inj + self.width)
        return np.where(inside, self.level, 0)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# pylint: disable=missing-module-docstring

import importlib.metadata

# Import from the gap submodule
from .gap import GapMaskGenerator

# Import from the glitch submodule
from .glitch.base import INJECTION_POINTS, Glitch
from .glitch.lpf import (
    IntegratedOneSidedDoubleExpGlitch,
    IntegratedShapeletGlitch,
    IntegratedTwoSidedDoubleExpGlitch,
    LPFLibraryGlitch,
    LPFLibraryModelGlitch,
    OneSidedDoubleExpGlitch,
    ShapeletGlitch,
    TwoSidedDoubleExpGlitch,
)
from .glitch.math import FunctionGlitch, RectangleGlitch, StepGlitch
from .glitch.read import HDF5Glitch, TimeSeriesGlitch

# Automatically set by `poetry dynamic-versioning`
__version__ = "2.0.0"

try:
    metadata = importlib.metadata.metadata("lisaglitch").json
    __author__ = metadata["author"]
    __email__ = metadata["author_email"]
except importlib.metadata.PackageNotFoundError:
    pass



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# pylint: disable=all
"""
Neural Spline Flow.

References:
    Paper reference: Durkan et al. <https://arxiv.org/abs/1906.04032>
    Code reference: slightly modified from <https://github.com/tonyduan/normalizing-flows/blob/master/nf/flows.py>
    Based on original implementation from <https://github.com/bayesiains/nsf>
"""

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from .resnet import ResidualNet

DEFAULT_MIN_BIN_WIDTH = 1e-3
DEFAULT_MIN_BIN_HEIGHT = 1e-3
DEFAULT_MIN_DERIVATIVE = 1e-3


class MLP(nn.Module):
    """Define a simple fully connected 4-layer MLP."""

    def __init__(self, number_in, number_out, number_hidden):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(number_in, number_hidden),
            nn.ReLU(),
            nn.Linear(number_hidden, number_hidden),
            nn.ReLU(),
            nn.Linear(number_hidden, number_hidden),
            nn.ReLU(),
            nn.Linear(number_hidden, number_out),
            nn.ReLU(),
        )

    def forward(self, x_input):
        """Forward pass for training."""
        return self.net(x_input)


def searchsorted(bin_locations, inputs, eps=1e-6):
    """Search."""
    bin_locations[..., -1] += eps
    return torch.sum(inputs[..., None] >= bin_locations, dim=-1) - 1


def unconstrained_RQS(
    inputs,
    unnormalized_widths,
    unnormalized_heights,
    unnormalized_derivatives,
    inverse=False,
    tail_bound=1.0,
    min_bin_width=DEFAULT_MIN_BIN_WIDTH,
    min_bin_height=DEFAULT_MIN_BIN_HEIGHT,
    min_derivative=DEFAULT_MIN_DERIVATIVE,
):
    """Unconstrained Rational Quadratic Spline."""

    inside_intvl_mask = (inputs >= -tail_bound) & (inputs <= tail_bound)
    outside_interval_mask = ~inside_intvl_mask

    outputs = torch.zeros_like(inputs)
    logabsdet = torch.zeros_like(inputs)

    unnormalized_derivatives = F.pad(unnormalized_derivatives, pad=(1, 1))
    constant = np.log(np.exp(1 - min_derivative) - 1)
    unnormalized_derivatives[..., 0] = constant
    unnormalized_derivatives[..., -1] = constant

    outputs[outside_interval_mask] = inputs[outside_interval_mask]
    logabsdet[outside_interval_mask] = 0

    if inside_intvl_mask is False:
        raise ValueError("Input outside domain")

    if torch.any(inside_intvl_mask):
        outputs[inside_intvl_mask], logabsdet[inside_intvl_mask] = RQS(
            inputs=inputs[inside_intvl_mask],
            unnormalized_widths=unnormalized_widths[inside_intvl_mask, :],
            unnormalized_heights=unnormalized_heights[inside_intvl_mask, :],
            unnormalized_derivatives=unnormalized_derivatives[inside_intvl_mask, :],
            inverse=inverse,
            left=-tail_bound,
            right=tail_bound,
            bottom=-tail_bound,
            top=tail_bound,
            min_bin_width=min_bin_width,
            min_bin_height=min_bin_height,
            min_derivative=min_derivative,
        )
    return outputs, logabsdet


def RQS(
    inputs,
    unnormalized_widths,
    unnormalized_heights,
    unnormalized_derivatives,
    inverse=False,
    left=0.0,
    right=1.0,
    bottom=0.0,
    top=1.0,
    min_bin_width=DEFAULT_MIN_BIN_WIDTH,
    min_bin_height=DEFAULT_MIN_BIN_HEIGHT,
    min_derivative=DEFAULT_MIN_DERIVATIVE,
):
    """Rational Quadratic Spline."""

    if torch.min(inputs) < left or torch.max(inputs) > right:
        raise ValueError("Input outside domain")

    num_bins = unnormalized_widths.shape[-1]

    if min_bin_width * num_bins > 1.0:
        raise ValueError("Minimal bin width too large for the number of bins")
    if min_bin_height * num_bins > 1.0:
        raise ValueError("Minimal bin height too large for the number of bins")

    widths = F.softmax(unnormalized_widths, dim=-1)
    widths = min_bin_width + (1 - min_bin_width * num_bins) * widths
    cumwidths = torch.cumsum(widths, dim=-1)
    cumwidths = F.pad(cumwidths, pad=(1, 0), mode="constant", value=0.0)
    cumwidths = (right - left) * cumwidths + left
    cumwidths[..., 0] = left
    cumwidths[..., -1] = right
    widths = cumwidths[..., 1:] - cumwidths[..., :-1]

    derivatives = min_derivative + F.softplus(unnormalized_derivatives)

    heights = F.softmax(unnormalized_heights, dim=-1)
    heights = min_bin_height + (1 - min_bin_height * num_bins) * heights
    cumheights = torch.cumsum(heights, dim=-1)
    cumheights = F.pad(cumheights, pad=(1, 0), mode="constant", value=0.0)
    cumheights = (top - bottom) * cumheights + bottom
    cumheights[..., 0] = bottom
    cumheights[..., -1] = top
    heights = cumheights[..., 1:] - cumheights[..., :-1]

    if inverse:
        bin_idx = searchsorted(cumheights, inputs)[..., None]
    else:
        bin_idx = searchsorted(cumwidths, inputs)[..., None]

    input_cumwidths = cumwidths.gather(-1, bin_idx)[..., 0]
    input_bin_widths = widths.gather(-1, bin_idx)[..., 0]

    input_cumheights = cumheights.gather(-1, bin_idx)[..., 0]
    delta = heights / widths
    input_delta = delta.gather(-1, bin_idx)[..., 0]

    input_derivatives = derivatives.gather(-1, bin_idx)[..., 0]
    input_derivatives_plus_one = derivatives[..., 1:].gather(-1, bin_idx)
    input_derivatives_plus_one = input_derivatives_plus_one[..., 0]

    input_heights = heights.gather(-1, bin_idx)[..., 0]

    if inverse:
        a = (inputs - input_cumheights) * (
            input_derivatives + input_derivatives_plus_one - 2 * input_delta
        ) + input_heights * (input_delta - input_derivatives)
        b = input_heights * input_derivatives - (inputs - input_cumheights) * (
            input_derivatives + input_derivatives_plus_one - 2 * input_delta
        )
        c = -input_delta * (inputs - input_cumheights)

        discriminant = b.pow(2) - 4 * a * c
        assert (discriminant >= 0).all()

        root = (2 * c) / (-b - torch.sqrt(discriminant))
        outputs = root * input_bin_widths + input_cumwidths

        theta_one_minus_theta = root * (1 - root)
        denominator = input_delta + (
            (input_derivatives + input_derivatives_plus_one - 2 * input_delta)
            * theta_one_minus_theta
        )
        derivative_numerator = input_delta.pow(2) * (
            input_derivatives_plus_one * root.pow(2)
            + 2 * input_delta * theta_one_minus_theta
            + input_derivatives * (1 - root).pow(2)
        )
        logabsdet = torch.log(derivative_numerator) - 2 * torch.log(denominator)
        return outputs, -logabsdet
    else:
        theta = (inputs - input_cumwidths) / input_bin_widths
        theta_one_minus_theta = theta * (1 - theta)

        numerator = input_heights * (
            input_delta * theta.pow(2) + input_derivatives * theta_one_minus_theta
        )
        denominator = input_delta + (
            (input_derivatives + input_derivatives_plus_one - 2 * input_delta)
            * theta_one_minus_theta
        )
        outputs = input_cumheights + numerator / denominator

        derivative_numerator = input_delta.pow(2) * (
            input_derivatives_plus_one * theta.pow(2)
            + 2 * input_delta * theta_one_minus_theta
            + input_derivatives * (1 - theta).pow(2)
        )
        logabsdet = torch.log(derivative_numerator) - 2 * torch.log(denominator)
        return outputs, logabsdet


class NSF_CL(nn.Module):
    """Neural spline flow, coupling layer from [1].

    [1] Durkan et al. (2019)
    """

    def __init__(self, dim, K=5, B=3, hidden_dim=8, base_network=ResidualNet):
        super().__init__()
        self.dim = dim
        self.K = K
        self.B = B
        self.f1 = base_network(dim // 2, (3 * K - 1) * dim // 2, hidden_dim)
        self.f2 = base_network(dim // 2, (3 * K - 1) * dim // 2, hidden_dim)

    def forward(self, x):
        """Forward pass."""
        log_det = torch.zeros(x.shape[0])
        lower, upper = x[:, : self.dim // 2], x[:, self.dim // 2 :]
        out = self.f1(lower).reshape(-1, self.dim // 2, 3 * self.K - 1)
        W, H, D = torch.split(out, self.K, dim=2)
        W, H = torch.softmax(W, dim=2), torch.softmax(H, dim=2)
        W, H = 2 * self.B * W, 2 * self.B * H
        D = F.softplus(D)
        upper, ld = unconstrained_RQS(upper, W, H, D, inverse=False, tail_bound=self.B)
        log_det += torch.sum(ld, dim=1)
        out = self.f2(upper).reshape(-1, self.dim // 2, 3 * self.K - 1)
        W, H, D = torch.split(out, self.K, dim=2)
        W, H = torch.softmax(W, dim=2), torch.softmax(H, dim=2)
        W, H = 2 * self.B * W, 2 * self.B * H
        D = F.softplus(D)
        lower, ld = unconstrained_RQS(lower, W, H, D, inverse=False, tail_bound=self.B)
        log_det += torch.sum(ld, dim=1)
        return torch.cat([lower, upper], dim=1), log_det

    def backward(self, z):
        """Backward pass to sample."""
        log_det = torch.zeros(z.shape[0])
        lower, upper = z[:, : self.dim // 2], z[:, self.dim // 2 :]
        out = self.f2(upper).reshape(-1, self.dim // 2, 3 * self.K - 1)
        W, H, D = torch.split(out, self.K, dim=2)
        W, H = torch.softmax(W, dim=2), torch.softmax(H, dim=2)
        W, H = 2 * self.B * W, 2 * self.B * H
        D = F.softplus(D)
        lower, ld = unconstrained_RQS(lower, W, H, D, inverse=True, tail_bound=self.B)
        log_det += torch.sum(ld, dim=1)
        out = self.f1(lower).reshape(-1, self.dim // 2, 3 * self.K - 1)
        W, H, D = torch.split(out, self.K, dim=2)
        W, H = torch.softmax(W, dim=2), torch.softmax(H, dim=2)
        W, H = 2 * self.B * W, 2 * self.B * H
        D = F.softplus(D)
        upper, ld = unconstrained_RQS(upper, W, H, D, inverse=True, tail_bound=self.B)
        log_det += torch.sum(ld, dim=1)
        return torch.cat([lower, upper], dim=1), log_det



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BSD 3-Clause License
#
# Copyright 2022, by the California Institute of Technology.
# ALL RIGHTS RESERVED. United States Government Sponsorship acknowledged.
# Any commercial use must be negotiated with the Office of Technology Transfer
# at the California Institute of Technology.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
# pylint: disable=C0103
"""
Based on https://github.com/karpathy/pytorch-normalizing-flows
"""

import torch
from torch import nn


class AffineConstantFlow(nn.Module):
    """
    Scales + Shifts the flow by (learned) constants per dimension.
    """

    def __init__(self, dim, scale=True, shift=True):
        super().__init__()
        self.s = (
            nn.Parameter(torch.randn(1, dim, requires_grad=True)) if scale else None
        )
        self.t = (
            nn.Parameter(torch.randn(1, dim, requires_grad=True)) if shift else None
        )

    def forward(self, x):
        """
        Forward pass for training.
        """
        s = self.s if self.s is not None else x.new_zeros(x.size())
        t = self.t if self.t is not None else x.new_zeros(x.size())
        z = x * torch.exp(s) + t
        log_det = torch.sum(s, dim=1)
        return z, log_det

    def backward(self, z):
        """
        Backward pass for sampling.
        """
        s = self.s if self.s is not None else z.new_zeros(z.size())
        t = self.t if self.t is not None else z.new_zeros(z.size())
        x = (z - t) * torch.exp(-s)
        log_det = torch.sum(-s, dim=1)
        return x, log_det


class ActNorm(AffineConstantFlow):
    """
    As described in Glow paper (https://arxiv.org/abs/1807.03039).
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_dep_init_done = False

    def forward(self, x):
        """
        Forward pass for training.
        """
        # first batch is used for init
        if not self.data_dep_init_done:
            assert self.s is not None and self.t is not None  # for now
            self.s.data = (-torch.log(x.std(dim=0, keepdim=True))).detach()
            self.t.data = (-(x * torch.exp(self.s)).mean(dim=0, keepdim=True)).detach()
            self.data_dep_init_done = True
        return super().forward(x)


class NormalizingFlow(nn.Module):
    """A sequence of Normalizing Flows is a Normalizing Flow"""

    def __init__(self, flows):
        super().__init__()
        self.flows = nn.ModuleList(flows)

    def forward(self, x):
        """
        Forward pass for training.
        """
        m, _ = x.shape
        log_det = torch.zeros(m)
        zs = [x]
        for flow in self.flows:
            x, ld = flow.forward(x)
            log_det += ld
            zs.append(x)
        return zs, log_det

    def backward(self, z):
        """
        Backward pass for sampling.
        """
        m, _ = z.shape
        log_det = torch.zeros(m)
        xs = [z]
        for flow in self.flows[::-1]:
            z, ld = flow.backward(z)
            log_det += ld
            xs.append(z)
        return xs, log_det


class NormalizingFlowModel(nn.Module):
    """A Normalizing Flow Model is a (prior, flow) pair"""

    def __init__(self, prior, flows):
        super().__init__()
        self.prior = prior
        self.flow = NormalizingFlow(flows)

    def forward(self, x):
        """
        Forward pass for training.
        """
        zs, log_det = self.flow.forward(x)
        prior_logprob = self.prior.log_prob(zs[-1]).view(x.size(0), -1).sum(1)
        return zs, prior_logprob, log_det

    def backward(self, z):
        """
        Backward pass for sampling.
        """
        xs, log_det = self.flow.backward(z)
        return xs, log_det

    def sample(self, num_samples):
        """
        Backward pass for sampling. Sample n=num_samples at a time.
        """
        z = self.prior.sample((num_samples,))
        xs, _ = self.flow.backward(z)
        return xs


class Invertible1x1Conv(nn.Module):
    """
    As introduced in Glow paper (https://arxiv.org/abs/1807.03039).
    """

    def __init__(self, dim):
        super().__init__()
        self.dim = dim
        Q = torch.nn.init.orthogonal_(torch.randn(dim, dim))
        P, L, U = torch.lu_unpack(*Q.lu())
        self.P = P  # remains fixed during optimization
        self.L = nn.Parameter(L)  # lower triangular portion
        self.S = nn.Parameter(U.diag())  # "crop out" the diagonal to its own parameter
        self.U = nn.Parameter(
            torch.triu(U, diagonal=1)
        )  # "crop out" diagonal, stored in S

    def _assemble_W(self):
        """
        Assemble W from its pieces (P, L, U, S)
        """
        L = torch.tril(self.L, diagonal=-1) + torch.diag(torch.ones(self.dim))
        U = torch.triu(self.U, diagonal=1)
        W = self.P @ L @ (U + torch.diag(self.S))
        return W

    def forward(self, x):
        """
        Forward pass for training.
        """
        W = self._assemble_W()
        z = x @ W
        log_det = torch.sum(torch.log(torch.abs(self.S)))
        return z, log_det

    def backward(self, z):
        """
        Backward pass for sampling.
        """
        W = self._assemble_W()
        W_inv = torch.inverse(W)
        x = z @ W_inv
        log_det = -torch.sum(torch.log(torch.abs(self.S)))
        return x, log_det



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BSD 3-Clause License
#
# Copyright 2022, by the California Institute of Technology.
# ALL RIGHTS RESERVED. United States Government Sponsorship acknowledged.
# Any commercial use must be negotiated with the Office of Technology Transfer
# at the California Institute of Technology.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Models used to fit glitches in LISA Pathfinder data, and associated libraries.

Authors:
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
    Eleonora Castelli <eleonora.castelli@unitn.it>
    Natalia Korsakova <natalia.korsakova@obspm.fr>
"""

import logging
import random
from typing import Any

import h5py
import numpy as np
from scipy.special import eval_genlaguerre  # pylint: disable=no-name-in-module

from .base import Glitch
from .read import HDF5Glitch

logger = logging.getLogger(__name__)


def _safe_exp(x):
    """Safely compute exponential, suppressing overflow warnings."""
    with np.errstate(over="ignore", invalid="ignore"):
        return np.exp(x)


def _safe_subtract(a, b):
    """Safely subtract arrays, suppressing invalid operation warnings."""
    with np.errstate(invalid="ignore"):
        return a - b


def _safe_multiply(a, b):
    """Safely multiply arrays, suppressing overflow warnings."""
    with np.errstate(over="ignore", invalid="ignore"):
        return a * b


class OneSidedDoubleExpGlitch(Glitch):
    r"""One-sided double-exponential glitch (LPF) [m/s2].

    A one-sided double-exponential glitch begins at  begins at
    :math:`t_\text{inj}`, ramps up to an amplitude :math:`A / (t_\text{rise} -
    t_\text{fall})` via the :math:`t_\text{rise}` timescale, and flattens off
    via the :math:`t_\text{fall}` timescale to 0.

    Defining :math:`\delta t = t - t_\text{inj}`, the signal is given by

    .. math::

        \frac{A (e^{-\delta t / t_\text{rise}} - e^{-\delta t /
        t_\text{fall}})}{t_\text{rise} - t_\text{fall}},

    and converges to 0 when :math:`t` goes to infinity.

    When :math:`t_\text{fise} = t_\text{fall}`, the previous equation becomes
    singular. We use the following continuous extension,

    .. math::

        \frac{A \delta t  e^{-\delta t / t_\text{fall}}}{t_\text{fall}^2}.

    Args:
        t_rise: rise timescale [s]
        t_fall: fall timescale [s]
        level: amplitude [injection point unit]
    """

    def __init__(
        self, t_rise: float, t_fall: float, level: float = 1.0, **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.t_rise = float(t_rise)  #: Rise timescale [s].
        self.t_fall = float(t_fall)  #: Fall timescale [s].
        self.level = float(level)  #: Amplitude [injection point unit].

    def compute_signal(self, t: np.ndarray) -> np.ndarray:
        delta_t = t - self.t_inj
        if self.t_fall != self.t_rise:
            dbexp = _safe_subtract(
                _safe_exp(-delta_t / self.t_rise), _safe_exp(-delta_t / self.t_fall)
            )
            y_inside = self.level * dbexp / (self.t_rise - self.t_fall)
        else:
            y_inside = (
                self.level
                * delta_t
                * _safe_exp(-delta_t / self.t_fall)
                / self.t_fall**2
            )
        return np.where(delta_t >= 0, y_inside, 0)


class IntegratedOneSidedDoubleExpGlitch(Glitch):
    r"""Integrated one-sided double-exponential glitch (LPF) [m/s].

    The integrated signal implemented here is given by

    .. math::

        A \qty[1 - \frac{t_\text{rise} e^{-\delta t / t_\text{rise}} -
        t_\text{fall} e^{-\delta t / t_\text{fall}}}{t_\text{rise} -
        t_\text{fall}}],

    and converges to :math:`A` when :math:`t` goes to infinity.

    When :math:`t_\text{rise} = t_\text{fall} \neq 0`, the previous expression
    becomes singular. We use use the following continuous extension,

    .. math::

        A \qty[1 - \qty(1 + \frac{\delta t}{t_\text{fall}}) e^{-\delta t /
        t_\text{fall}}].

    Args:
        t_rise (float): rise timescale [s]
        t_fall (float): fall timescale [s]
        level (float): amplitude [injection point unit]
    """

    def __init__(
        self, t_rise: float, t_fall: float, level: float = 1.0, **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.t_rise = float(t_rise)  #: Rise timescale [s].
        self.t_fall = float(t_fall)  #: Fall timescale [s].
        self.level = float(level)  #: Amplitude [injection point unit].

    def compute_signal(self, t: np.ndarray) -> np.ndarray:
        delta_t = t - self.t_inj
        if self.t_fall != self.t_rise:
            dbexp = _safe_subtract(
                _safe_multiply(self.t_rise, _safe_exp(-delta_t / self.t_rise)),
                _safe_multiply(self.t_fall, _safe_exp(-delta_t / self.t_fall)),
            )
            y_inside = self.level * (1 - dbexp / (self.t_rise - self.t_fall))
        else:
            y_inside = self.level * (
                1 - (1 + delta_t / self.t_fall) * _safe_exp(-delta_t / self.t_fall)
            )
        return np.where(delta_t >= 0, y_inside, 0)


class TwoSidedDoubleExpGlitch(Glitch):
    r"""Integrated two-sided double-exponential glitch (LPF) [m/s2].

    A two-sided double-exponential glitch begins at :math:`t_\text{inj}`, ramps
    up to level :math:`A` via the :math:`t_\text{rise}` timescale, oscillates
    around zero and then and flattens off via the :math: `t_\text{fall}`
    timescale to 0.

    Defining :math:`\delta t = t - t_\text{inj}`, the signal is given by

    .. math::

        \frac{1}{t_\text{fall} - t_\text{rise}} \qty[\frac{D + t_\text{fall}
        A}{t_\text{rise}} e^{-\delta t / t_\text{rise}} - \frac{D +
        t_\text{rise} A}{t_\text{fall}} e^{-\delta t / t_\text{fall}}],

    and falls back to 0 when :math:`t` goes to infinity.

    When :math:`t_\text{fise} = t_\text{fall}`, the previous equation becomes
    singular. We use the following continuous extension,

    .. math::

        A \qty[1 - \qty(1 - \frac{\delta t}{t_\text{fall}} - \delta t \frac{D}{A
        t_\text{fall}^2}) e^{-\delta t / t_\text{fall}}].

    Args:
        t_rise (float): rise timescale [s]
        t_fall (float): fall timescale [s]
        level (float): amplitude [injection point unit]
        displacement (float): induced displacement [m]
    """

    def __init__(
        self,
        t_rise: float,
        t_fall: float,
        level: float = 1.0,
        displacement: float = 1.0,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.t_rise = float(t_rise)  #: Rise timescale [s].
        self.t_fall = float(t_fall)  #: Fall timescale [s].
        self.level = float(level)  #: Amplitude [injection point unit].
        self.displacement = float(displacement)  #: Induced displacement [m].

    def compute_signal(self, t: np.ndarray) -> np.ndarray:
        delta_t = t - self.t_inj
        if self.t_fall != self.t_rise:
            dbexp = (
                self.displacement + self.level * self.t_fall
            ) / self.t_rise * np.exp(-delta_t / self.t_rise) - (
                self.displacement + self.level * self.t_rise
            ) / self.t_fall * np.exp(
                -delta_t / self.t_fall
            )
            y_inside = dbexp / (self.t_rise - self.t_fall)
        else:
            dbexp = -delta_t * (
                self.displacement + self.level * self.t_fall
            ) + self.t_fall * (self.displacement + 2 * self.level * self.t_fall)
            y_inside = dbexp * np.exp(-delta_t / self.t_fall) / self.t_fall**3
        return np.where(delta_t >= 0, y_inside, 0)


class IntegratedTwoSidedDoubleExpGlitch(Glitch):
    r"""Integrated two-sided double-exponential glitch (LPF) [m/s].

    The integrated signal implemented here is given by

    .. math::

        A + \frac{\qty[e^{-\delta t / t_\text{rise}} (D + A t_\text{fall}) -
        e^{-\delta t / t_\text{fall}} (D + A t_\text{rise})]}{t_\text{fall} -
        t_\text{rise}},

    and falls back to 0 when :math:`t` goes to infinity.

    When :math:`t_\text{fise} = t_\text{fall}`, the previous equation becomes
    singular. We use the following continuous extension,

    .. math::

        A \qty[1 - \qty(1 - \frac{\delta t}{t_\text{fall}} - \frac{D \delta t}{A
        t_\text{fall}^2}) e^{-\delta t / t_\text{fall}}].

    Args:
        t_rise (float): rise timescale [s]
        t_fall (float): fall timescale [s]
        level (float): amplitude [injection point unit]
        displacement (float): induced displacement [m]
    """

    def __init__(
        self,
        t_rise: float,
        t_fall: float,
        level: float = 1.0,
        displacement: float = 1.0,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.t_rise = float(t_rise)  #: Rise timescale [s].
        self.t_fall = float(t_fall)  #: Fall timescale [s].
        self.level = float(level)  #: Amplitude [injection point unit].
        self.displacement = float(displacement)  #: Induced displacement [m].

    def compute_signal(self, t: np.ndarray) -> np.ndarray:
        delta_t = t - self.t_inj
        if self.t_fall != self.t_rise:
            dbexp = (self.displacement + self.level * self.t_fall) * np.exp(
                -delta_t / self.t_rise
            ) - (self.displacement + self.level * self.t_rise) * np.exp(
                -delta_t / self.t_fall
            )
            y_inside = self.level + dbexp / (self.t_rise - self.t_fall)
        else:
            y_inside = self.level * (
                1
                - (
                    1
                    - delta_t / self.t_fall
                    - delta_t * self.displacement / (self.level * self.t_fall**2)
                )
                * np.exp(-delta_t / self.t_fall)
            )
        return np.where(delta_t >= 0, y_inside, 0)


class ShapeletGlitch(Glitch):
    r"""Exponential shapelet glitch. [m/s2]

    The glitch signal is given by the normalized 1D hydrogen atom wavefunctions.
    For the default case ``n = 1``, the signal is given by

    .. math::

        2 \alpha \frac{\delta t}{\beta} e^{-\delta t / \beta},

    where :math:`\alpha = \frac{A}{\beta}`, with :math:`A` the glitch amplitude
    and :math:`\beta` the glitch damping time, :math:`\delta t = t - t_\text{inj}`.

    For other values of ``n``,

    .. math::

        2 \alpha (-1)^{n-1} \frac{\delta t}{\beta} n^{-2.5} L^1_{n-1} \frac{\delta
        t}{n \beta} e^{-\delta t / (n \beta)},

    where :math:`L^1_{n-1}(t)` is the generalized Laguerre polynomial.

    This model was used to fit the LISA Pathfinder glitches and produce estimate
    of the the glitch distribution, as well as the production of LDC data.

    This model is equivalent to OneSidedDoubleExpGlitch with

    .. math::

        \lim_{t_\text{rise} \to t_\text{fall}} t_\text{fall} = \beta

    and

    .. math::

        \Delta v = 2 \alpha \beta

    Args:
        level (float): amplitude [injection point unit]
        beta (float): damping time [s]
        quantum_n (int): number of shapelet components (quantum energy level)
    """

    def __init__(
        self, level: float = 1, beta: float = 1.0, quantum_n: int = 1, **kwargs
    ) -> None:
        super().__init__(**kwargs)

        self.level = float(level)  #: Amplitude [injection point unit].
        self.beta = float(beta)  #: Damping time [s].
        self.quantum_n = int(
            quantum_n
        )  #: Number of shapelet components (quantum energy level).

    def compute_signal(self, t: np.ndarray) -> np.ndarray:
        alpha = self.level / self.beta / 2
        delta_t = t - self.t_inj
        t_inside = delta_t >= 0
        normalized_t = 2 * delta_t[t_inside] / (self.quantum_n * self.beta)

        # Scipy special functions are written in C, and are not
        # correctly recognized by Pylint, so disabling linting here
        laguerre = eval_genlaguerre(self.quantum_n - 1, 1, normalized_t)

        y_inside = np.zeros_like(t)
        y_inside[t_inside] = (
            2
            * alpha
            * (-1) ** (self.quantum_n - 1)
            * delta_t[t_inside]
            / self.beta
            * self.quantum_n ** (-2.5)
            * laguerre
            * np.exp(-normalized_t / 2)
        )

        return y_inside


class IntegratedShapeletGlitch(Glitch):
    r"""Single-component (``n = 1``) integrated exponential shapelet glitch. [m/s]

    The glitch signal is given by the normalized 1D hydrogen atom wavefunctions,
    which we integrate over time with ``n = 1``,

    .. math::

        2 \alpha \beta \qty[1 - \qty(1 + \frac{\delta t}{\beta}) e^{-
        \delta t / \beta}].

    where :math:`\alpha = \frac{A}{\beta}`, with :math:`A` the glitch amplitude
    and :math:`\beta` the glitch damping time, :math:`\delta t = t - t_\text{inj}`.

    This model was used to fit the LISA Pathfinder glitches and produce estimate
    of the the glitch distribution, as well as the production of LDC data.

    This model is equivalent to OneSidedDoubleExpGlitch with

    .. math::

        \lim_{t_\text{rise} \to t_\text{fall}} t_\text{fall} = \beta

    and

    .. math::

        \Delta v = 2 \alpha \beta

    Args:
        level (float): amplitude [injection point unit]
        beta (float): damping time [s]
    """

    def __init__(self, level: float = 1.0, beta: float = 1.0, **kwargs) -> None:
        super().__init__(**kwargs)
        self.level = float(level)  #: Amplitude [injection point unit].
        self.beta = float(beta)  #: Damping time [s].

    def compute_signal(self, t: np.ndarray) -> np.ndarray:
        alpha = self.level / self.beta / 2
        delta_t = t - self.t_inj
        t_inside = delta_t >= 0
        normalized_t = delta_t[t_inside] / self.beta

        result = np.zeros_like(t)
        result[t_inside] = (
            2 * alpha * self.beta * (1 - (1 + normalized_t) * np.exp(-normalized_t))
        )
        return result


class LPFLibraryGlitch(HDF5Glitch):
    """Glitch from a LISA Pathfinder glitch library.

    The LISA Pathfinder glitch library has the following structure::

        timeseries / runXX / glitchYY

    You can specify a run and glitch number, or leave them as ``None`` to let
    the class pick a random run and/or glitch.

    The latest LPF glitch library is available for download in the LISA Glitch
    project's `data repository
    <https://gitlab.in2p3.fr/lisa-simulation/glitch/-/tree/master/data>`_. The
    library provides acceleration signals, used by default by this class. If you
    wish to inject velocity glitch, use `integrated=True` to use an analytic
    time-integrated version of the signal.

    Args:
        path (str): path to LPF glitch library
        run (int or None): run number, or None to sample a random run
        glitch (int or None): glitch number, or None to sample a random glitch
        integrated (bool): whether to integrate glitch signal (use a velocity glitch)
    """

    def __init__(
        self,
        path: str,
        run: int | None = None,
        glitch: int | None = None,
        integrated: bool = False,
        **kwargs,
    ) -> None:

        self.library = str(path)  #: Path to LPF glitch library.
        logger.info("Opening LPF glitch library '%s'", self.library)
        hdf5 = h5py.File(self.library, "r")

        # Pick run
        if run is None:
            self.run = None
            logger.debug("Picking a random run in library '%s'", self.library)
            timeseries_group = hdf5.get("timeseries")
            assert isinstance(timeseries_group, h5py.Group)
            random_group = random.choice(list(timeseries_group.keys()))
            self.run_group = f"timeseries/{random_group}"
        else:
            self.run = int(run)
            logger.debug("Using user-provided run '%s'", self.run)
            self.run_group = f"timeseries/run{self.run:02}"

        try:
            run_group_object = hdf5[self.run_group]
        except KeyError as error:
            raise KeyError(
                f"run group '{self.run_group}' not found in '{self.library}'"
            ) from error
        assert isinstance(run_group_object, h5py.Group)

        # Pick glitch
        if glitch is None:
            self.glitch = None
            logger.debug("Picking a random glitch in run '%s'", self.run)
            random_dataset = random.choice(list(run_group_object.keys()))
            self.glitch_dataset = f"{self.run_group}/{random_dataset}"
        else:
            self.glitch = int(glitch)
            logger.debug("Using user-provided glitch '%s'", self.glitch)
            self.glitch_dataset = (
                f"{self.run_group}/glitch{self.glitch:02}"  #: Path to glitch dataset.
            )

        try:
            hdf5[self.glitch_dataset]
        except KeyError as error:
            raise KeyError(
                f"glitch dataset '{self.glitch_dataset}' not found in '{self.library}'"
            ) from error

        logger.info("Closing LPF glitch library '%s'", self.library)
        hdf5.close()

        super().__init__(self.library, self.glitch_dataset, **kwargs)

        self.integrated = bool(
            integrated
        )  #: Whether to integrate glitch signal (use a velocity glitch).
        if self.integrated:
            # Integrate interpolant once analytically
            logger.debug(
                "Integrating LPF library acceleration glitch signal to velocity"
            )
            self.interpolant = self.interpolant.antiderivative()


class LPFLibraryModelGlitch:
    r"""Double-exponential glitch using parameters from a LPF glitch library.

    Glitch parameters are read from the library, and either
    :class:`lisaglitch.OneSidedDoubleExpGlitch` or
    :class:`lisaglitch.TwoSidedDoubleExpGlitch` class is used to generate the
    signal.

    The LISA Pathfinder glitch library has the following structure::

        legacy_params / runXX / glitchYY

    You can specify a run and glitch number, or leave them as ``None`` to let
    the class pick a random run and/or glitch.

    The latest LPF glitch library is available for download in the LISA Glitch
    project's `data repository
    <https://gitlab.in2p3.fr/lisa-simulation/glitch/-/tree/master/data>`_. The
    libnrary provides acceleration signals, used by default by this class. If
    you wish to inject velocity glitch, use `integrated=True` to use an analytic
    time-integrated version of the signal.

    Args:
        path (str): path to LPF glitch library
        run (int or None): run number, or None to sample a random run
        glitch (int or None): glitch number, or None to sample a random glitch
        integrated (bool): whether to integrate glitch signal (use a velocity glitch)

    Attributes:
        library: Path to LPF glitch library.
        run_group: Path to run group.
        glitch_dataset: Path to glitch dataset.
        integrated: Whether to integrate glitch signal (use a velocity glitch).
    """

    def __new__(
        cls,
        path: str,
        run: int | None = None,
        glitch: int | None = None,
        integrated: bool = False,
        **kwargs,
    ) -> Any:

        library = str(path)
        logger.info("Opening LPF glitch library '%s'", library)
        hdf5 = h5py.File(library, "r")

        # Pick run
        if run is None:
            run = None
            logger.debug("Picking a random run in library '%s'", library)
            legacy_params_group = hdf5.get("legacy_params")
            assert isinstance(legacy_params_group, h5py.Group)
            random_group = random.choice(list(legacy_params_group.keys()))
            run_group = f"legacy_params/{random_group}"
        else:
            run = int(run)
            logger.debug("Using user-provided run '%s'", run)
            run_group = f"legacy_params/run{run:02}"

        try:
            run_group_object = hdf5[run_group]
        except KeyError as error:
            raise KeyError(
                f"run group '{run_group}' not found in '{library}'"
            ) from error
        assert isinstance(run_group_object, h5py.Group)

        # Pick glitch
        if glitch is None:
            glitch = None
            logger.debug("Picking a random glitch in run '%s'", run)
            random_dataset = random.choice(list(run_group_object.keys()))
            glitch_dataset = f"{run_group}/{random_dataset}"
        else:
            glitch = int(glitch)
            logger.debug("Using user-provided glitch '%s'", glitch)
            glitch_dataset = f"{run_group}/glitch{glitch:02}"

        try:
            legacy_params_ds = hdf5[glitch_dataset]
            assert isinstance(legacy_params_ds, h5py.Dataset)
            legacy_params = legacy_params_ds[:]
        except KeyError as error:
            raise KeyError(
                f"glitch dataset '{glitch_dataset}' not found in '{library}'"
            ) from error

        logger.info("Closing LPF glitch library '%s'", library)
        hdf5.close()

        # Retreive parameters
        level = float(legacy_params[0])
        t_rise = float(legacy_params[2])
        t_fall = float(legacy_params[3])
        displacement = float(legacy_params[4])

        if not integrated:
            # Initialize the acceleration glitch
            if displacement != 0:
                instance: Any = TwoSidedDoubleExpGlitch(
                    t_rise, t_fall, level, displacement, **kwargs
                )
            else:
                instance = OneSidedDoubleExpGlitch(t_rise, t_fall, level, **kwargs)
        else:
            # Initialize the (integrated) velocity glitch
            if displacement != 0:
                instance = IntegratedTwoSidedDoubleExpGlitch(
                    t_rise, t_fall, level, displacement, **kwargs
                )
            else:
                instance = IntegratedOneSidedDoubleExpGlitch(
                    t_rise, t_fall, level, **kwargs
                )

        # Set parameters as attributes on the glitch instance
        instance.library = library
        instance.integrated = integrated
        instance.run = run
        instance.run_group = run_group
        instance.glitch = glitch
        instance.glitch_dataset = glitch_dataset
        instance.legacy_params = legacy_params
        instance.plot = lambda *args, **kwargs: LPFLibraryModelGlitch.plot(
            instance, *args, **kwargs
        )

        return instance

    @staticmethod
    def plot(
        instance: "LPFLibraryModelGlitch",
        *args,
        **kwargs,
    ) -> None:
        """Plot glitch signal.

        See :meth:`lisaglitch.Glitch.plot`.
        """
        instance.__class__.plot(instance, *args, **kwargs)



"""
LISA gap mask generation submodule.

Provides classes and utilities for creating realistic data gap patterns
that simulate mission interruptions and data quality issues.
"""

from .mask import GapMaskGenerator



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BSD 3-Clause License
#
# Copyright 2022, by the California Institute of Technology.
# ALL RIGHTS RESERVED. United States Government Sponsorship acknowledged.
# Any commercial use must be negotiated with the Office of Technology Transfer
# at the California Institute of Technology.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Helper functions to build a LPF distribution and sample glitches from it.

Authors:
    Natalia Korsakova <natalia.korsakova@obspm.fr>
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
"""

import itertools
import logging

import numpy as np
import torch
from torch import optim
from torch.distributions import MultivariateNormal

from .flow import ActNorm, Invertible1x1Conv, NormalizingFlowModel
from .nsflow import NSF_CL

logger = logging.getLogger(__name__)


def initialise_network(dev):
    """Initialize network architecture that will be used for training and sampling.

    Initialize prior distribution, the flow and optimizer.

    Args:
        dev: device to use ('cuda:0' or 'cpu')

    Returns:
        2-tuple (mode, optimizer).
    """
    prior = MultivariateNormal(torch.zeros(2), torch.eye(2))

    # Neural splines, coupling
    nfs_flow = NSF_CL
    flows = [nfs_flow(dim=2, K=8, B=3, hidden_dim=256) for _ in range(9)]
    convs = [Invertible1x1Conv(dim=2) for _ in flows]
    norms = [ActNorm(dim=2) for _ in flows]
    flows = list(itertools.chain(*zip(norms, convs, flows)))

    # Construct the model
    model = NormalizingFlowModel(prior, flows).to(dev)
    # Optimizer
    optimizer = optim.Adam(model.parameters(), lr=5e-5, weight_decay=1e-5)

    return model, optimizer


class LPFGlitchParameterSampler:
    """Sample glitch parameters in the LISA Pathfinder distribution."""

    def __init__(self, dev, min_beta, max_beta, min_amp, max_amp):
        """Initialize a sampler for LPG glitch parameters.

        Args:
            dev: device to use ('cuda:0' or 'cpu')
            min_beta: minimum Poisson parameter for renomarlization
            max_beta: maximum Poisson parameter for renomarlization
            min_amp: minimum glitch amplitude for renomarlization
            max_amp: maximum glitch amplitude for renomarlization
        """

        self.model_path = None
        self.model = None
        self.optimizer = None
        self.dev = dev

        # Parameters to renormalise back to the proper values.
        self.min_beta = min_beta
        self.max_beta = max_beta
        self.min_amp = min_amp
        self.max_amp = max_amp

    def define_model(self):
        """Define the model for which the weights are loaded."""
        self.model, self.optimizer = initialise_network(self.dev)

    def load_model(self, model_path):
        """Load trained model from the file.

        Args:
            model_path: path to model file
        """
        self.model_path = model_path
        checkpoint = torch.load(self.model_path, map_location=torch.device(self.dev))
        self.model.load_state_dict(checkpoint["model_state_dict"])
        self.optimizer.load_state_dict(checkpoint["optimizer_state_dict"])

    def sample_param(self, number_samples):
        """Sample the parameters from the fitted weights.

        Args:
            number_samples: number of points to sample
        """
        zsample = self.model.sample(number_samples)
        z_point = zsample[-1]
        z_point = z_point.detach().cpu().numpy()
        return z_point

    def generator(self, beta_range_max=None, amp_range_min=None):
        """Generator that outputs the next pair of the glitch parameters.

        Note that `beta_range` and `amp_range` are conditions that can be set
        additionally if we want to impose some restriction on the
        glitch distribution (like `beta_range` < 150, `amp_range` > 1e-12).

        Args:
            beta_range_max: maximum allowed Poisson parameter
            amp_range_min: minimum allowed glitch amplitude
        """
        beta = 0.0
        amp = 0.0

        number_samples = 1
        while True:

            lpf_samples = self.sample_param(number_samples)

            beta_correct_range = np.exp(
                self.min_beta + lpf_samples[:, 0] * (self.max_beta - self.min_beta)
            )
            amp_correct_range = np.exp(
                self.min_amp + lpf_samples[:, 1] * (self.max_amp - self.min_amp)
            )

            if beta_range_max is not None and amp_range_min is not None:
                if (
                    beta_correct_range < beta_range_max
                    and amp_correct_range > amp_range_min
                ):
                    beta = beta_correct_range[0]
                    amp = amp_correct_range[0]
                    yield beta, amp
                else:
                    continue
            else:
                beta = beta_correct_range[0]
                amp = amp_correct_range[0]
                yield beta, amp

    def generate(self, number_samples):
        """Generate many samples at once.

        Args:
            number_samples: number of samples
        """
        lpf_samples = self.sample_param(number_samples)
        beta_correct_range = np.exp(
            self.min_beta + lpf_samples[:, 0] * (self.max_beta - self.min_beta)
        )
        amp_correct_range = np.exp(
            self.min_amp + lpf_samples[:, 1] * (self.max_amp - self.min_amp)
        )

        return beta_correct_range, amp_correct_range


def estimate_poisson_beta(path_intervals):
    """Estimate Poisson's lambda parameter from interarrivals.

    Read a file containing glitch arrival time differences, and estimate the Poisson
    distribution's lambda parameter.

    Args:
        path_intervals: path to interarrival file

    Returns:
        Sample estimate of Poisson's lambda parameter.
    """
    interarrivals = np.loadtxt(
        path_intervals,
        comments="#",
        delimiter=" ",
        skiprows=0,
        usecols=(0),
        unpack=False,
    )
    return 1.0 / np.mean(interarrivals)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BSD 3-Clause License
#
# Copyright 2022, by the California Institute of Technology.
# ALL RIGHTS RESERVED. United States Government Sponsorship acknowledged.
# Any commercial use must be negotiated with the Office of Technology Transfer
# at the California Institute of Technology.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Base class for glitches

All glitch classes inherit from this base class.

Authors:
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
    Eleonora Castelli <eleonora.castelli@unitn.it>
"""

import abc
import importlib
import logging
from typing import Optional, Tuple

import h5py
import matplotlib.pyplot as plt
import numpy as np

logger = logging.getLogger(__name__)

INJECTION_POINTS = [
    # Test-mass motion
    "tm_12",
    "tm_23",
    "tm_31",
    "tm_13",
    "tm_32",
    "tm_21",
    # Laser frequency jumps
    "laser_12",
    "laser_23",
    "laser_31",
    "laser_13",
    "laser_32",
    "laser_21",
    # Readout of carrier inter-spacecraft interferometers
    "readout_isi_carrier_12",
    "readout_isi_carrier_23",
    "readout_isi_carrier_31",
    "readout_isi_carrier_13",
    "readout_isi_carrier_32",
    "readout_isi_carrier_21",
    # Readout of upper sideband inter-spacecraft interferometers
    "readout_isi_usb_12",
    "readout_isi_usb_23",
    "readout_isi_usb_31",
    "readout_isi_usb_13",
    "readout_isi_usb_32",
    "readout_isi_usb_21",
    # Readout of carrier test-mass interferometers
    "readout_tmi_carrier_12",
    "readout_tmi_carrier_23",
    "readout_tmi_carrier_31",
    "readout_tmi_carrier_13",
    "readout_tmi_carrier_32",
    "readout_tmi_carrier_21",
    # Readout of upper sideband test-mass interferometers
    "readout_tmi_usb_12",
    "readout_tmi_usb_23",
    "readout_tmi_usb_31",
    "readout_tmi_usb_13",
    "readout_tmi_usb_32",
    "readout_tmi_usb_21",
    # Readout of carrier reference interferometers
    "readout_rfi_carrier_12",
    "readout_rfi_carrier_23",
    "readout_rfi_carrier_31",
    "readout_rfi_carrier_13",
    "readout_rfi_carrier_32",
    "readout_rfi_carrier_21",
    # Readout of upper sideband reference interferometers
    "readout_rfi_usb_12",
    "readout_rfi_usb_23",
    "readout_rfi_usb_31",
    "readout_rfi_usb_13",
    "readout_rfi_usb_32",
    "readout_rfi_usb_21",
]


class Glitch(abc.ABC):
    """Abstract base class to represent a single glitch signal.

    A glitch is a time-domain signal which is injected into the instrumental
    simulation at a given injection point and at a given time This abstract
    class provides a common structure for all glitch classes.

    .. admonition:: Units and time coordinates

        The physical unit and time coordinate associated with a glitch are
        defined by the associated injection point.

    Args:
        inj_point (str): injection point, see
            :const:`lisaglitch.INJECTION_POINTS` for a list of valid injection
            points
        t_inj (float): injection time [s]
    """

    def __init__(self, inj_point, t_inj=0) -> None:

        self.git_url = "https://gitlab.in2p3.fr/lisa-simulation/glitch"
        self.generator = self.__class__.__name__
        self.version = importlib.metadata.version("lisaglitch")
        logger.info("Initializing glitch (lisaglitch version %s)", self.version)

        if inj_point not in INJECTION_POINTS:
            raise ValueError(f"Invalid injection point '{inj_point}'")
        self.inj_point = inj_point  #: Injection point.
        self.t_inj = float(t_inj)  #: Injection time [s].

    @abc.abstractmethod
    def compute_signal(self, t) -> np.ndarray:
        """Compute the glitch time-domain signal.

        The time is expressed in the local coordinates of the injection. E.g.,
        glitches injected in the test-mass onboard spacecraft 1 are expressed in
        the spacecraft proper time (TPS) of spacecraft 1.

        If ``t`` is None, we use the default sampling parameters of the glitch
        object.

        Args:
            t ((N,) array-like): time [s]

        Return:
            Time-domain glitch signal, of shape (N,).
        """
        raise NotImplementedError

    def plot(self, t, output=None, title="Glitch signal") -> None:
        """Plot glitch signal.

        Args:
            t ((N,) array-like): time [s]
            output (str or None): output file, None to show the plots
            title (str): plot title
        """
        # Compute glitch signal
        logger.info("Computing glitch signal")
        signal = self.compute_signal(t)
        # Create a new figure for this plot
        plt.figure(figsize=(10, 6))
        # Plot glitch
        logger.info("Plotting glitch signal")
        plt.plot(t, signal)
        plt.xlabel("Time [s]")
        plt.ylabel("Signal")
        plt.title(title)
        plt.grid(True, alpha=0.3)
        # Save or show glitch
        if output is not None:
            logger.info("Saving plot to %s", output)
            plt.savefig(output, bbox_inches="tight", dpi=150)
            plt.close()  # Close the figure to free memory
        else:
            plt.show()

    def _write_metadata(self, hdf5, prefix=""):
        """Set all properties as HDF5 attributes on ``object``.

        Try to store all variables as attributes. If it is too large or its type
        is not supported, try to store a string representation; if this fails,
        log a warning.

        Args:
            hdf5 (h5py.File): an HDF5 file, or a dataset
            prefix (str): prefix for attribute names
        """
        for key, value in self.__dict__.items():
            pkey = f"{prefix}{key}"
            try:
                hdf5.attrs[pkey] = value
            except (TypeError, RuntimeError):
                try:
                    hdf5.attrs[pkey] = str(value)
                except RuntimeError:
                    logger.warning("Cannot write metadata '%s' on '%s'", pkey, hdf5)

    @staticmethod
    def read_file_sampling(path: str) -> Optional[Tuple[float, int, float]]:
        """Read the sampling parameters from a glitch file.

        Args:
            path (str): path to the glitch file

        Returns:
            ``(dt, size, t0)`` for the glitch file, or ``None`` if the file does
            does not exist or does not contain the sampling parameters.
        """
        logger.info("Reading sampling parameters from glitch file '%s'", path)
        try:
            with h5py.File(path, "r") as hdf5:
                dt = hdf5.attrs["dt"]
                assert isinstance(dt, (float, np.floating))
                size = hdf5.attrs["size"]
                assert isinstance(size, (int, np.integer))
                t0 = hdf5.attrs["t0"]
                assert isinstance(t0, (float, np.floating))
        except (FileNotFoundError, KeyError):
            logger.info("Cannot read sampling parameters from '%s'", path)
            return None

        return float(dt), int(size), float(t0)

    def write(self, path, dt=0.0625, size=2592000, t0=0.0, mode="a") -> None:
        """Write the glitch to a glitch file.

        If the file does not exist, it is created with a time axis matching
        ``dt``, ``size``, and ``t0`` arguments. The glitch signal is computed
        according to these parameters and written to file.

        If the file already exists, we make sure that ``dt``, ``size``, and
        ``t0`` match the values used to create the file and raise an error if
        they do not. Use :meth:`lisaglitch.Glitch.read_file_sampling` to get the
        value for these parameters from an existing file.

        When creating the glitch file, metadata are saved as attributes.

        When writing a glitch, we add attributes for each local variable on the
        associated injection point dataset, prefixed with ``inj<i>``, where i is
        the index of glitch in the injection point dataset.

        Args:
            path (str): path to the glitch file
            dt (float): time step [s]
            size (int): number of samples
            t0 (float): start time [s]
            mode (str): opening mode

        Raises:
            ValueError: if the file already exists and the sampling parameters
            do not match the values used to create the file
        """
        # Open orbit file
        with h5py.File(path, mode) as hdf5:
            # Write sampling parameters if needed
            sampling = self.read_file_sampling(path)
            if sampling is None:
                logger.info("New glitch file, writing sampling parameters")
                hdf5.attrs["dt"] = dt
                hdf5.attrs["size"] = size
                hdf5.attrs["t0"] = t0
                logger.info("Setting global metadata")
                self._write_metadata(hdf5)
            else:
                logger.info("Existing glitch file, checking sampling parameters")
                if sampling != (dt, size, t0):
                    raise ValueError("Sampling parameters do not match")
            # Create injection dataset if needed
            t = t0 + np.arange(size, dtype=np.float64) * dt
            dname = self.inj_point
            if dname not in hdf5:
                logger.info("Creating injection dataset '%s'", dname)
                hdf5.create_dataset(dname, data=np.zeros_like(t))
                hdf5[dname].attrs["injection_count"] = 0
            # Add glitch to injection dataset
            logger.info("Adding glitch to injection dataset '%s'", dname)
            ninj = int(hdf5[dname].attrs["injection_count"])
            hdf5[dname][:] += self.compute_signal(t)
            # Setting metadata
            logger.info("Setting injection metadata")
            hdf5[dname].attrs["injection_count"] = ninj + 1
            self._write_metadata(hdf5[dname], prefix=f"inj{ninj}_")
        # Closing file
        logger.info("Closing glitch file %s", path)
